import re

from rocolib.utils.numsym import Dummy
import logging
import yaml


### XXX "." is used to separate subcomponent parameters when inherited
###     but we can't tell the difference here, so it's not invalid I guess
VALID_ATTR = re.compile('^[a-zA-Z_][a-zA-Z0-9_.]*$')

PARAM_TYPES = {
    "length": {
        "valueType": "float",
        "minValue": 0,
        "units": "mm",
    },
    "angle": {
        "valueType": "float",
        "minValue": 0,
        "maxValue": 360,
        "units": "degrees",
    },
    "ratio": {
        "valueType": "float",
        "minValue": 0,
        "maxValue": 1,
    },
    "count": {
        "valueType": "int",
        "minValue": 0,
    },
}

class Parameter:
    def __init__(self, name, defaultValue=None, paramType=None, **kwargs):
        if not VALID_ATTR.match(name):
            raise ValueError("Invalid parameter name " + name)
        if name[0] == '_' and not kwargs.get("hidden", False):
            raise ValueError("Leading _ character reserved for hidden parameters, either set hidden flag or change name from " + name)

        self.name = name

        if defaultValue is None and not kwargs.get("overrides", False):
            raise ValueError(f"Must specify either defaultValue or overrides (implies optional) for parameter {name}")

        self.defaultValue = defaultValue
        self.spec = {}

        if paramType in PARAM_TYPES:
            for k, v in PARAM_TYPES[paramType].items():
                self.spec[k] = v
        elif paramType:
            raise ValueError(f"Unknown paramType: {paramType}")
        for k, v in kwargs.items():
            self.spec[k] = v
        self.assertValid(defaultValue)

        vt = self.spec.get("valueType", "")
        no = not(self.spec.get("overrides", False))

        if no and ("int" in vt or "float" in vt):
            integer=None
            positive=None

            if "float" not in vt:
                integer=True

            mv = self.spec.get("minValue", -1)
            if mv is not None and mv >= 0:
                positive=True

            self.symbol = Dummy(name, real=True, positive=positive, integer=integer)
        else:
            self.symbol = None
        self.value = None

    def resolveValue(self, value):
        if value is None:
            return None
        vt = self.spec.get("valueType")
        if vt:
            value = eval(vt)(value)
        if "values" not in self.spec:
            return value
        sanitize = lambda x : yaml.safe_load(yaml.safe_dump(x))
        vtest = sanitize(value)
        v = sanitize(self.spec["values"])
        if isinstance(v, dict) and vtest not in list(v.keys()):
            ks = list(v.keys())
            vs = list(v.values())
            if vtest in vs:
                return ks[vs.index(vtest)]
        return value

    def setValue(self, value, validate=True):
        value = self.resolveValue(value)
        if validate:
            self.assertValid(value)
        self.value = value

    def setDefault(self, force=False):
        if force or self.value is None:
            self.setValue(self.defaultValue)

    def setSpec(self, spec, value, validate=False):
        self.spec[spec] = value
        if validate:
            self.assertValid(self.value)

    def assertValid(self, value):
        if value is None:
            if self.spec.get("overrides", False):
                return True
            else:
                raise ValueError(f"Parameter {self.name} is not optional (via overrides) and cannot be set to None")

        def check(spec, test, error):
            if (self.spec.get(spec, None) is not None and not test(self.spec[spec])):
                raise ValueError(f"When setting parameter {self.name}: {value} {error} {self.spec[spec]}")

        check("valueType", lambda x : eval(x)(value) is not None, "is not of type")
        vt = self.spec.get("valueType")
        if vt:
            value = eval(vt)(value)
        check("minValue", lambda x : value >= x, "is less than")
        check("maxValue", lambda x : value <= x, "is greater than")
        check("values", lambda x : value in x, "is not in")
        check("isValid", lambda x : x(self, value), "is invalid")
        return True

    def validate(self):
        return self.assertValid(self.value)

    def getValue(self):
        if self.value is not None:
            if "values" in self.spec and isinstance(self.spec["values"], dict):
                return self.spec["values"][self.value]
            else:
                return self.value
        else:
            return self.symbol

    def doOverrides(self, c):
        if self.value is None:
            return
        overrides = self.spec.get("overrides", [])
        if isinstance(overrides, dict):
            for k, fn in overrides.items():
                try:
                    import rocolib.utils.numsym as np
                    value = eval(fn, locals())
                    c.setParameter(k, value)
                except ValueError as e:
                    raise ValueError(str(e) + f" (via setting overriding parameter {self.name} to {self.value})")
                except KeyError as e:
                    logging.debug(str(e) + f" (via setting overriding parameter {self.name} to {self.value}) -- ignoring")
        elif isinstance(overrides, (list, tuple)):
            for k in overrides:
                try:
                    c.setParameter(k, self.value)
                except KeyError as e:
                    logging.debug(str(e) + f" (via setting overriding parameter {self.name} to {self.value}) -- ignoring")
        else:
            try:
                c.setParameter(overrides, self.value)
            except KeyError as e:
                logging.debug(str(e) + f" (via setting overriding parameter {self.name} to {self.value}) -- ignoring")

    def getInfo(self, keys=None):
        if isinstance(keys, str):
            return getattr(self, keys)
        elif keys:
            return {k: v for k, v in self.__dict__.items() if k[0] != '_' and k in keys}
        else:
            return {k: v for k, v in self.__dict__.items() if k[0] != '_'}

    def getSpec(self):
        return self.spec

class Parameterized(object):
    """
    Like a dictionary k/v store, but we require special syntax constructs
    to set/update keys
    """
    def __init__(self, name = None):
        self._parameters = {}
        self.name = name
        class P: pass
        self.p = P()

    @property
    def name(self):
      return self._name if self._name is not None else f"C-{str(self.__class__)}"

    @name.setter
    def name(self, name):
      self._name = name

    def addParameter(self, name, defaultValue=None, **kwargs):
        """
        Adds a k/v pair to the internal store if the key has not been added before
        Raises KeyError if the key has been added before or if the parameter name is already a class property/method
        """
        if name in self._parameters:
            raise KeyError("Parameter %s already exists on object %s" % (name, str(self)))
        if name in dir(self.p):
            raise KeyError("Parameter name %s is a reserved string for object %s" % (name, str(self)))
        p = Parameter(name, defaultValue, **kwargs)
        self._parameters.setdefault(name, p)
        setattr(self.p.__class__, name, property(
            lambda s: self.getParameter(name),
            lambda s, v: self.setParameter(name, v),
            lambda s: self.delParameter(name)
        ))
        return p


    def useDefaultParameters(self, force=False):
        """
        Assign the default value to all parameters:
        - Only if otherwise unset if force=False
        - Overwrite any previously set values if force=True
        """
        for n, p in self._parameters.items():
            p.setDefault(force)
        for n, p in self._parameters.items():
            p.doOverrides(self)


    def setParameter(self, n, v, validate=True):
        """
        Sets a k/v pair to the internal store if the key has been added previously
        Raises KeyError if the key has not been added before
        Passes along any ValueErrors from Parameter object
        """
        if "#" in n:
            n, spec = n.split("#")
        else:
            spec = None
        if n in self._parameters:
            if spec:
                self._parameters[n].setSpec(spec, v, validate)
            else:
                self._parameters[n].setValue(v, validate)
        else:
            raise KeyError("Parameter %s not does not exist on object %s" % (n, str(self)))


    def validate(self):
        for p in self._parameters.values():
            p.validate()

    def getParameter(self, name):
        """
        Retrieves the parameter value with the given name
        Raises KeyError if the key is not been set
        """
        return self._parameters[name].getValue()


    def getParameterInfo(self, name=None, keys=None, hidden=False):
        """
        Retrieves the parameter metadata info
        """
        if name:
            return self._parameters[name].getInfo(keys)
        else:
            return {k: v.getInfo(keys) for k, v in self._parameters.items() if hidden or not v.getSpec().get("hidden", False)}


    def hasParameter(self, name):
        return name in self._parameters


    def delParameter(self, name):
        if name in self._parameters:
            self._parameters.pop(name)
            delattr(self.p.__class__, name)
        else:
            raise KeyError("Parameter %s not does not exist on object %s" % (name, str(self)))
