from rocolib.api.composables.Composable import Composable, output, widget
import networkx as nx
import json
from circlify import circlify
from rocolib.utils.nx2go import GraphVisualization as gv
from dash import dcc, html, Input, Output, ctx, ALL
from dash.exceptions import PreventUpdate
import dash_bootstrap_components as dbc
import logging
from pprint import pformat


class ComponentComposable(Composable):
    def __init__(self, component):
        self.c = component
        self.net = nx.DiGraph()
        self.net.graph['depth'] = 0

    def append(self, c2, name, **kwargs):
        sc = c2.c
        scname = sc.name
        scclass = sc.cName
        scnet = c2.net
        
        self.net.add_node(scname, name=name, component=scclass, net=scnet)
        self.net.graph['depth'] = max(self.net.graph['depth'], scnet.graph['depth']+1)

    def attach(self, fromPort, toPort, **kwargs):
        self.net.add_edge(fromPort.parent.name, toPort.parent.name)

    def _makeComponentMap(self, scd=None, full=None, idnum = 0):
        if scd is None:
            g = self.net
            full = nx.DiGraph()
            full.add_node(0, name="", component=self.c.name)
            title = f"<b>{self.c.name}</b>"
        else:
            try:
                g = scd['net']
            except KeyError:
                print(scd, idnum)
                raise
            title = f"<b>{scd['component']}</b><br>{scd['name']}"

        data = dict(id=idnum,
                    datum=(g.graph['depth']+1)**2,
                    children=[dict(id=len(full), datum=0.5)])
        full.add_node(len(full), title=title)

        if g:
            f = nx.convert_node_labels_to_integers(g, len(full))
            full = nx.disjoint_union(full, f)
            for sc, scd in f.nodes(data=True):
                full, cd = self._makeComponentMap(scd, full, sc)
                data["children"].append(cd)
        return full, data

    def _drawComponentTree(self):
        g, data = self._makeComponentMap()
        circles = circlify([data])

        tree = gv(g,
            pos = {c.ex['id']: (c.x, c.y) for c in circles},
            node_size = {c.ex['id']: c.r for c in circles},
            node_text = lambda k, v : v.get("title", ""),
            node_border_width = lambda k, v: "title" not in v and 2 or 0,
            edge_width = 0,
        )
        fig = tree.create_figure(width=1024, height=768)
        return fig

    @output("map.png", "component topology map", binary=True)
    def drawComponentMap(self, fp, **ka):
        fig = self._drawComponentTree()
        fig.write_image(fp, format="png")

    @widget('displaymap', "component topology map")
    def displayMap(self, fp, **ka):
        return dcc.Graph(figure = self._drawComponentTree())

    @widget('parameters', "component parameter form")
    def paramForm(self, fp, **ka):
        inputs = []

        inputs.append(html.H4("Interfaces"))
        inputrow = []
        for i, (name, info) in enumerate(self.c.getInterfaceInfo().items()):
            port=info.get("port", "")

            disabled = False
            if port=="EdgePort":
                color = 'primary'
            elif port=="FacePort":
                color = 'info'
            else:
                disabled = True
                color = 'secondary'

            elem = dbc.Button(name, n_clicks=0,
                id=dict(name=name, type='interface', port=port),
                color = color,
                disabled = disabled,
            )
            inputrow.append( dbc.Col ( html.Div(elem, className="d-grid gap-2") , width=2) )
            if i % 6 == 5:
                inputs.append(dbc.Row(inputrow))
                inputrow = []
        if inputrow:
            inputs.append(dbc.Row(inputrow))

        inputs.append(html.H4("Parameters"))
        for name, info in self.c.getParameterInfo().items():

            val = info.get("value", info.get("defaultValue", None))
            if val is None:
                val = ''

            inputrow = [ dbc.Col( html.Div(name), width=3 ) , dbc.Col( html.Div(json.dumps(val)), width=1 ) ]

            spec = info.get("spec", {})
            elem = dbc.Input
            idict = dict(
                id = dict(name=name, type='parameter', valueType=("values" in spec and 'values' or spec.get("valueType", ''))),
                type = 'text',
                value = val,
                debounce = True,
            )

            minval = maxval = None
            if "minValue" in spec and spec["minValue"] is not None:
                minval = idict["min"] = spec.get("minValue")
                minval = round(minval, 2)
            if "maxValue" in spec and spec["maxValue"] is not None:
                maxval = idict["max"] = spec.get("maxValue")
                minval = round(maxval, 2)

            if spec.get("values"):
                elem = dbc.Select
                idict.pop("type")
                idict.pop("debounce")
                idict["options"] = [dict(label=x, value=json.dumps(x)) for x in spec.get("values")]
                idict["value"] = json.dumps(val)
            elif spec.get("valueType") == "int":
                idict["type"] = 'number'
            elif spec.get("valueType") == "float":
                idict["type"] = 'range'
                if "max" not in idict:
                    mx = int(max(val * 2, idict.get("min", 0) + 10))
                    idict["max"] = mx
                    maxval = f"{mx} >"
                if "min" not in idict:
                    idict["min"] = int(min(val - 10, 0))
                    minval = f"< {idict['min']}"
                delta = idict['max'] - idict['min']
                idict["step"] = 0.1
                if delta > 5:
                    idict["step"] = 1
                elif delta > 20:
                    idict["step"] = 2
                elif delta > 100:
                    idict["step"] = 5
            elif spec.get("valueType") == "bool":
                elem = dbc.Switch
                idict.pop("type")
                idict.pop("debounce")
                idict["label"] = name
                idict["value"] = val
            else:
                logging.info(f"Unknown valueType: {spec.get('valueType')} for parameter {name}, using default text input")
                idict["value"] = json.dumps(val)

            inputgroup = []

            if minval is not None:
                inputgroup.append( dbc.InputGroupText(minval) )

            e = elem(**idict)
            inputgroup.append( e )

            if maxval is not None:
                inputgroup.append( dbc.InputGroupText(maxval) )

            inputrow.append( dbc.Col(dbc.InputGroup(inputgroup)) )
            inputs.append(dbc.Row(inputrow))

        inputs.append(html.Div("Hello world", id="output_params"))

        app = ka.pop("app", None)
        if app:
            @app.callback(Output(dict(composable=ALL, name=ALL, type="widget"), "children"),
                          Input(dict(name=ALL, type="parameter", valueType=ALL), "value"),
                          Input(dict(name=ALL, type="interface", port=ALL), "n_clicks"))
            def updateParams(values, btns):
                if not ctx.triggered_id:
                    raise PreventUpdate
                tname = ctx.triggered_id["name"]
                ttype = ctx.triggered_id["type"]

                outputs = [ d["id"]["name"] for d in ctx.outputs_list ]
                if ttype == "parameter":
                    params = _extract(ctx.inputs_list[0])
                    self.c.reload(params)
                    ka.pop("remake", None)
                elif ttype == "interface":
                    ka["remake"] = False
                    iface = self.c.getInterface(tname)
                    if ctx.triggered_id["port"] == "EdgePort":
                        ka["highlight"] = iface.edges
                    elif ctx.triggered_id["port"] == "FacePort":
                        ka["highlight"] = [e.name for e in iface.getFace().edges]
                    else:
                        raise PreventUpdate
                widgets = self.c.makeOutput(outputs, widgets=True, **ka)
                return [widgets[d["id"]["composable"]][d["id"]["name"]]['widget'] for d in ctx.outputs_list]

            @app.callback(Output('output_params', 'children'),
                Input(dict(name=ALL, type="parameter", valueType=ALL), "value"))
            def updateParams(values):
                if not ctx.triggered_id:
                    s = f"Current parameters set to:"
                else:
                    s = f"Parameter changed: {ctx.triggered_id}; rebuilding with new parameters:"
                params = _extract(ctx.inputs_list[0])
                return html.Div([html.Div(s), html.Code(pformat(params))])

        return html.Div(dbc.Container(inputs))

def _extract(ctxlist):
    params = {}
    for i in ctxlist:
        val = i['value']
        if val == '':
            val = None
        elif i['id']['valueType'] not in ('int', 'float', 'bool'):
            try:
                val = json.loads(val)
            except (TypeError, json.decoder.JSONDecodeError):
                logging.debug(f"Error json-parsing param {i['id']['name']} = {val} ({type(val)})")
                logging.debug(f"... leaving as string")
                pass
        params[i['id']['name']] = val
    return params
