from rocolib.api.components import FoldedComponent
from rocolib.api.composables.graph.Face import Face, Rectangle
from rocolib.utils.numsym import sqrt, arccos, rad2deg

class PinchedRectBeam(FoldedComponent):
  def define(self):
    self.addParameter("topwidth", 10, paramType="length")
    self.addParameter("topdepth", 30, paramType="length")
    self.addParameter("botwidth", 20, paramType="length")
    self.addParameter("botdepth", 20, paramType="length")
    self.addParameter("length", 30, paramType="length")

    for i in range(4):
        self.addEdgeInterface(f"topedge{i}", f"s{i}t.e2", ["topwidth", "topdepth"][i % 2])
        self.addEdgeInterface(f"botedge{i}", f"s{i}b.e2", ["botwidth", "botdepth"][i % 2])

    self.addEdgeInterface("top", [f"s{i}t.e2" for i in range(4)], ["topwidth", "topdepth"])
    self.addEdgeInterface("bot", [f"s{3-i}b.e2" for i in range(4)], ["botdepth", "botwidth"])

  def modifyParameters(self):
    tw = self.p.topwidth
    td = self.p.topdepth
    bw = self.p.botwidth
    bd = self.p.botdepth
    if (tw + td != bw + bd):
        raise ValueError(f"Unequal perimeters: top = {2*(tw+td)}, bot = {2*(bw+bd)}")

  def assemble(self):
    tw = self.p.topwidth
    td = self.p.topdepth
    bw = self.p.botwidth
    bd = self.p.botdepth
    l  = self.p.length

    if bw < tw:
        mirror = True
        bw, bd = bd, bw
        tw, td = td, tw
    else:
        mirror = False

    x = (bw-tw)/2
    a = 90-rad2deg(arccos(x/l))
    q = 180-rad2deg(arccos(x*x/(l*l)))

    rs = []
    rs.append(Face("", ((0, -l),  (x, 0),      (0,0))))
    rs.append(Face("", ((0, -l),  (bw, -l),    (bw - x, 0),  (x, 0))))
    rs.append(Face("", ((bw, -l), (bw+bd, -l), (bw+bd+x, 0), (bw-x, 0))))
    rs.append(Face("", ((0, -l),  (bw, -l),    (bw - x, 0),  (x, 0))))
    rs.append(Face("", ((bw, -l), (bw+bd, -l), (bw+bd, 0),   (bw-x, 0))))

    self.attachEdge(None, rs[0], "e2", prefix="r0", angle=90)
    self.attachEdge("r0.e1", rs[1], "e0", prefix="r1", angle=q)
    self.attachEdge("r1.e2", rs[2], "e0", prefix="r2", angle=q)
    self.attachEdge("r2.e2", rs[3], "e0", prefix="r3", angle=q)
    self.attachEdge("r3.e2", rs[4], "e0", prefix="r4", angle=q)

    self.attachEdge("r1.e1", Rectangle("", bw, 0), "e0", prefix="s0b", angle=-a)
    self.attachEdge("r2.e1", Rectangle("", bd, 0), "e0", prefix="s1b", angle=a)
    self.attachEdge("r3.e1", Rectangle("", bw, 0), "e0", prefix="s2b", angle=-a)
    self.attachEdge("r4.e1", Rectangle("", bd, 0), "e0", prefix="s3b", angle=a)

    self.attachEdge("r1.e3", Rectangle("", tw, 0), "e0", prefix="s0t", angle=a)
    self.attachEdge("r2.e3", Rectangle("", td, 0), "e0", prefix="s1t", angle=-a)
    self.attachEdge("r3.e3", Rectangle("", tw, 0), "e0", prefix="s2t", angle=a)
    self.attachEdge("r4.e3", Face("", ((td-x, 0), (td, 0), (0, 0))), "e0", prefix="s3t", angle=-a)

    self.addTab("r0.e0", "r4.e2", angle= 0, width=min(10, bd))

    if mirror:
        self.getGraph().mirror()
        for j in range(4):
            i = 3-j
            self.setEdgeInterface(f"topedge{j}", f"s{i}t.e2", ["topwidth", "topdepth"][i % 2])
            self.setEdgeInterface(f"botedge{j}", f"s{i}b.e2", ["botwidth", "botdepth"][i % 2])

if __name__ == "__main__":
  PinchedRectBeam.test()
