from rocolib.api.components import FoldedComponent
from rocolib.api.composables.graph.Face import Rectangle, Face
from rocolib.api.ports import AnchorPort
from rocolib.utils.utils import decorateGraph
from rocolib.utils.transforms import Translate, RotateZ
from rocolib.utils.numsym import dot, deg2rad


class ServoMotor(FoldedComponent):
    def define(self):
        self.addParameter("angle", 180, paramType="angle", minValue=None, maxValue=None)
        self.addParameter("centermount", True, valueType="bool")
        self.addParameter("hornmount", True, valueType="bool")
        self.addParameter("hornslots", True, valueType="bool")
        self.addParameter("flip", False, valueType="bool")

        self.addParameter("servo", "fs90r", values=ServoMotor.dims)
        self.addInterface("mount", AnchorPort(self, self.getGraph(), "horn", Translate([0,0,0])))
        self.addFaceInterface("horn", "horn")

    def assemble(self):
        s = self.p.servo

        dz = s.get("hornheight")
        dy = s.get("motorlength") / 2 - s.get("hornoffset")
        if self.p.flip:
            dy = -dy

        f = Rectangle("horn", 0, 0)
        if self.p.centermount:
            decorateGraph(f, Rectangle("centerhole", 1, 1))
        if self.p.hornmount:
            decorateGraph(f, Face("hornhole", [(ds*dd, s.get("hornlength")+dd) for dd in (-0.1, 0.1) for ds in (-1, 1)], recenter=False))
        if self.p.hornslots:
            x = s.get("hornwidth") + s.get("horndepth")
            dx = s.get("horndepth")*1.5 / 2
            dy = ((s.get("hornwidth") + s.get("horndepth")*.5) / 2) / dx
            while x < s.get("hornlength"):
                decorateGraph(f, Face(f"hornslotp{x}", [(ds*dd, x+dd) for dd in (-dx, dx) for ds in (-dy, dy)], recenter=False))
                decorateGraph(f, Face(f"hornslotn{x}", [(ds*dd, -x+dd) for dd in (-dx, dx) for ds in (-dy, dy)], recenter=False))
                x += s.get("horndepth") * 3
        self.addFace(f)
        self.setInterface("mount", AnchorPort(self, self.getGraph(), "horn", dot(Translate([0,-dy,dz]), RotateZ(deg2rad(self.p.angle)))))


    '''
    Servo dimension parameters:

                 |<-G->|
    ^   =o======v======o= { H (X)J
    E          _I_
    v ________| | |_____
      ^ |       |<-F->|<> D
      | |             |
      B | <--- A ---> |
      | | (X) C       |
      v |_____________|

    A : motorlength
    B : motorheight
    C : motorwidth
    D : shoulderlength

    E : hornheight
    F : hornoffset

    G : hornlength
    H : horndepth
    J : hornwidth

    '''

    dims = dict(
        ds2g = dict(
            motorlength   = 17,
            motorwidth    = 8.5,
            motorheight   = 11,
            shoulderlength= 3.5,
            hornlength    = 10,
            hornheight    = 9,
            hornoffset    = 4,
            horndepth     = 2,
            hornwidth     = 3,
        ),
        s4303r = dict(
            motorlength   = 31,
            motorwidth    = 17,
            motorheight   = 29,
            shoulderlength= 10,
            hornlength    = 38,
            hornheight    = 14,
            hornoffset    = 7,
            horndepth     = 2,
            hornwidth     = 3,
        ),
        tgy1370a = dict(
            motorlength   = 20,
            motorwidth    = 9,
            motorheight   = 14,
            shoulderlength= 4,
            hornlength    = 7,
            hornheight    = 10,
            hornoffset    = 4,
            horndepth     = 2,
            hornwidth     = 3,
        ),
        fs90r = dict(
            motorlength   = 23,
            motorwidth    = 13,
            motorheight   = 19,
            shoulderlength= 5,
            hornlength    = 10,
            hornheight    = 11,
            hornoffset    = 8,
            horndepth     = 2,
            hornwidth     = 5,
        ),
    )


if __name__ == "__main__":
    ServoMotor.test()
    print(ServoMotor.dims)
